"""Encode Python JSON-able objects as Cypher expressions."""


def encode_dict(obj):
    """Encode dictionary."""
    return "{" + ", ".join(
        key + ": " + dumps(value)
        for key, value in obj.items()
    ) + "}"


def encode_list(obj):
    """Encode list."""
    return "[" + ", ".join(
        dumps(el) for el in obj
    ) + "]"


def encode_str(obj):
    """Encode string."""
    return f"\"{obj}\""


def encode_none(obj):
    """Encode None."""
    return "null"


def encode_bool(obj):
    """Encode boolean."""
    return "true" if obj else "false"


def dumps(obj):
    """Convert Python obj to Cypher expression."""
    if isinstance(obj, dict):
        return encode_dict(obj)
    elif isinstance(obj, list):
        return encode_list(obj)
    elif isinstance(obj, str):
        return encode_str(obj)
    elif isinstance(obj, bool):
        return encode_bool(obj)
    elif obj is None:
        return encode_none(obj)
    else:
        return str(obj)



"""Query nesting tools."""
from functools import reduce
from operator import and_, or_


class Query():
    """Cypher query segment."""

    def __init__(
            self,
            string,
            qids=None,
            references=None,
            qgraph=None,
    ):
        """Initialize."""
        self._string = string
        self._qids = qids
        self._references = references
        self._qgraph = qgraph

    @property
    def qgraph(self):
        """Get qgraph."""
        return self._qgraph

    def logic(self, simple=True):
        """Return whether qid is required."""
        if simple:
            return ""
        conditions = [
            f"{qid} IS NOT null" for qid in self.qids
        ]
        if len(self.qids) > 1:
            conditions = [
                f"({condition})" for condition in conditions
            ]
        return " AND ".join(conditions)

    def compile(self, **kwargs):
        """Wrap hidden _compile method."""
        context = kwargs.get("context", set())
        ext_context = kwargs.get("ext_context", set())

        clauses = []
        imports = ext_context & self.references
        if imports:
            clauses.append("WITH " + ", ".join(imports))
        context = context | ext_context
        kwargs.pop("ext_context", None)
        kwargs["context"] = context

        return_ = kwargs.pop("return_", False)
        wrap = kwargs.pop("wrap", False)

        optional = kwargs.get("optional", False)
        extension = kwargs.pop("extension", [])

        if optional and not isinstance(self, CompoundQuery):
            if len(self.references) >= 5:
                # there are probably at least two edges here
                wrap = True
                kwargs.update(
                    optional=False,
                    extension=[(
                        "WITH CASE WHEN count(*) > 0 "
                        + "THEN collect([{0}]) ".format(
                            ", ".join(self.qids - context)
                        )
                        + "ELSE [[]] "
                        + "END AS results "
                        + "UNWIND results as result "
                        + "WITH {0}".format(", ".join(
                            f"result[{idx}] AS {qid}"
                            for idx, qid in enumerate(self.qids - context)
                        ))
                    )],
                )
            else:
                clauses.append("OPTIONAL")

        if wrap:
            clauses.extend(self._compile_wrapped(**kwargs))
        else:
            clauses.extend(self._compile(**kwargs))

        clauses.extend(extension)

        if return_:
            clauses.append(self.return_clause(**kwargs))

        return clauses

    def _compile_wrapped(self, **kwargs):
        """Compile wrapped query."""
        context = kwargs.pop("context", set())
        inner_context = context & self.references
        return [
            "CALL {{{query}}}".format(
                query=(
                    " ".join(self.compile(
                        ext_context=inner_context,
                        return_=True,
                        **kwargs,
                    ))
                    .replace("\\", "\\\\")
                ),
            ),
        ]

    def _compile(self, **kwargs):  # pylint: disable=unused-argument
        """Return query string."""
        return [self._string] if self._string else []

    @property
    def references(self):
        """Get references."""
        return self._references

    @property
    def qids(self):
        """Get qids."""
        return self._qids

    def where_clause(self):
        """Get WHERE clause."""
        conditions = self.logic()
        return conditions and "WHERE " + conditions

    def with_clause(self, **kwargs):
        """Get WITH clause."""
        context = kwargs.get("context", set())
        return "WITH " + ", ".join(self.qids | context)

    def return_clause(self, **kwargs):
        """Get RETURN clause."""
        context = kwargs.get("context", set())
        return "RETURN " + ", ".join(self.qids - context)

    def __and__(self, other):
        """AND two queries together."""
        return AndQuery(self, other)

    def __invert__(self):
        """NOT query."""
        return NotQuery(self)

    def __or__(self, other):
        """OR queries."""
        return OrQuery(self, other)

    def __xor__(self, other):
        """XOR queries."""
        return XorQuery(self, other)


class CompoundQuery(Query):
    """Compound query."""

    def __init__(self, *subqueries):  # pylint: disable=super-init-not-called
        """Initialize."""
        self.subqueries = subqueries

    @property
    def qids(self):
        """Get qids."""
        return reduce(
            or_,
            [subquery.qids for subquery in self.subqueries]
        )

    @property
    def qgraph(self):
        """Get qgraph."""
        qnodes = dict()
        qedges = dict()
        for subquery in self.subqueries:
            qnodes.update(subquery.qgraph["nodes"])
            qedges.update(subquery.qgraph["edges"])
        return {
            "nodes": qnodes,
            "edges": qedges,
        }

    @property
    def references(self):
        """Get references."""
        return reduce(
            or_,
            (query.references for query in self.subqueries),
        )


class AndQuery(CompoundQuery):
    """Compound query segment."""

    def _compile(self, **kwargs):
        """Get query string."""
        context = kwargs.pop("context", set())
        subquery_strings = []
        for query in self.subqueries:
            subquery_strings.extend(query.compile(context=context, **kwargs))
            context = context | query.qids

        return subquery_strings

    def logic(self, simple=True):
        """Return whether qid is required."""
        conditions = [
            query.logic(simple) for query in self.subqueries
            if query.logic(simple)
        ]
        if len(conditions) > 1:
            conditions = [
                f"({condition})" for condition in conditions
            ]
        return " AND ".join(conditions)


class NotQuery(CompoundQuery):
    """Not query segment."""

    def _compile(self, **kwargs):
        """Return query string."""
        kwargs.update(optional=True)
        return self.subqueries[0].compile(**kwargs)

    def logic(self, simple=True):
        """Return whether qid is required."""
        return f"NOT ({self.subqueries[0].logic(False)})"


class AltQuery(CompoundQuery):
    """Alternative query.

    Superclass for OrQuery and XorQuery.
    """

    def __init__(self, *args):
        """Initialize."""
        assert len(args) == 2
        super().__init__(*args)

    def _compile(self, **kwargs):
        """Get query string."""
        kwargs.update(optional=True)
        if self.subqueries[0].qids & self.subqueries[1].qids:
            return self._compile_union(**kwargs, return_=True)
        else:
            query = AndQuery(self.subqueries[0], self.subqueries[1])
            return query.compile(**kwargs)

    def _compile_union(self, **kwargs):
        """Get query string."""
        query0 = AndQuery(self.subqueries[0], self.subqueries[1])
        query1 = AndQuery(self.subqueries[1], self.subqueries[0])

        query = UnionQuery(query0, query1)
        return query.compile(wrap=True, **kwargs)


class UnionQuery(CompoundQuery):
    """UNION query."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        super().__init__(*args, **kwargs)
        self.return_ = None
        self.ext_context = None

    def compile(self, **kwargs):
        """Return query string."""
        assert kwargs.get("return_", False)

        # save these things and later pass them, unmodified, to subqueries
        self.return_ = kwargs.pop("return_", set())
        self.ext_context = kwargs.pop("ext_context", set())

        return super().compile(**kwargs)

    def _compile(self, **kwargs):
        """Return query string."""
        kwargs.update(
            ext_context=self.ext_context,
            return_=self.return_,
        )

        return [" UNION ".join(
            " ".join(subquery.compile(
                **kwargs,
            ))
            for subquery in self.subqueries
        )]


class OrQuery(AltQuery):
    """OR query."""

    def logic(self, simple=True):
        """Get conditions."""
        return " OR ".join(
            f"({query.logic(False)})" for query in self.subqueries
        )


class XorQuery(AltQuery):
    """XOR query."""

    def logic(self, simple=True):
        """Get conditions."""
        return " XOR ".join(
            f"({query.logic(False)})" for query in self.subqueries
        )



"""Setup reasoner-transpiler package."""
from setuptools import setup

setup(
    name="reasoner-transpiler",
    version="1.11.11",
    author="Patrick Wang",
    author_email="patrick@covar.com",
    url="https://github.com/ranking-agent/reasoner-transpiler",
    description="TRAPI → Cypher transpiler",
    packages=["reasoner_transpiler"],
    package_data={"reasoner_transpiler": ["attribute_types.json"]},
    install_requires=[
        "bmt-lite-v3.1.0>=2.2.2,<3.0",
    ],
    zip_safe=False,
    license="MIT",
    python_requires=">=3.6",
)



"""Utilities."""
import re
from typing import List, Union, overload


def _space_case(arg: str):
    """Convert string to space case.

    "ThisCase" is replaced with "this case".
    """
    # replace "_" with " "
    tmp = re.sub("_", " ", arg)
    # replace "xYz" with "x yz"
    tmp = re.sub(
        r"(?<=[a-z])(?=[A-Z][a-z])",
        " ",
        tmp
    )
    # lower-case everything
    return tmp.lower()


def _snake_case(arg: str):
    """Convert string to snake_case.

    "ThisCase" is replaced with "this_case".
    """
    # replace " " with "_"
    tmp = re.sub(
        " ", "_", arg,
    )
    # replace xYz with x_Yz
    tmp = re.sub(
        r"(?<=[a-z])(?=[A-Z][a-z])",
        "_",
        tmp
    )
    # lower-case everything
    return tmp.lower()


def _pascal_case(arg: str):
    """Convert string to PascalCase.

    "this_case" is replaced with "ThisCase".
    """
    # replace "_x" or " x" with "X"
    tmp = re.sub(
        r"(?<=[a-zA-Z])[ _]([a-z])",
        lambda c: c.group(1).upper(),
        arg
    )
    # upper-case first character
    tmp = re.sub(
        r"^[a-z]",
        lambda c: c.group(0).upper(),
        tmp
    )
    return tmp


@overload
def space_case(arg: str) -> str:
    """Convert a string to space case."""


@overload
def space_case(arg: List[str]) -> List[str]:
    """Convert a set of strings to space case."""


def space_case(arg: Union[str, List[str]]) -> Union[str, List[str]]:
    """Convert a string or set of strings to space case."""
    if isinstance(arg, str):
        return _space_case(arg)
    elif isinstance(arg, list):
        return [_space_case(arg) for arg in arg]
    else:
        raise ValueError()


def snake_case(arg: Union[str, List[str]]):
    """Convert each string or set of strings to snake_case."""
    if isinstance(arg, str):
        return _snake_case(arg)
    elif isinstance(arg, list):
        return [_snake_case(arg) for arg in arg]
    else:
        raise ValueError()


def pascal_case(arg: Union[str, List[str]]):
    """Convert each string or set of strings to PascalCase."""
    if isinstance(arg, str):
        return _pascal_case(arg)
    elif isinstance(arg, list):
        return [_pascal_case(arg) for arg in arg]
    else:
        raise ValueError()


def ensure_list(arg: Union[str, List[str]]) -> List[str]:
    """Convert scalar arg to a list of one."""
    if isinstance(arg, list):
        return arg
    return [arg]



"""Tools for compiling QGraph into Cypher query."""
import copy
from functools import reduce
import json
from operator import and_
from pathlib import Path

from . import cypher_expression
from .matching import match_query

DIR_PATH = Path(__file__).parent
with open(DIR_PATH / "attribute_types.json", "r") as stream:
    ATTRIBUTE_TYPES = json.load(stream)

RESERVED_NODE_PROPS = [
    "id",
    "category",
]
RESERVED_EDGE_PROPS = [
    "id",
    "predicate",
]

def nest_op(operator, *args):
    """Generate a nested set of operations from a flat expression."""
    if len(args) > 2:
        return [operator, args[0], nest_op(operator, *args[1:])]
    else:
        return [operator, *args]


def transpile_compound(qgraph, **kwargs):
    """Transpile compound qgraph."""
    if isinstance(qgraph, dict):
        return match_query(
            qgraph,
            **kwargs,
        )
    if qgraph[0] == "OR":
        qgraph = nest_op(*qgraph)

    args = [
        transpile_compound(arg, **kwargs)
        for arg in qgraph[1:]
    ]
    if qgraph[0] == "AND":
        return reduce(and_, args)
    elif qgraph[0] == "OR":
        return args[0] | args[1]
    elif qgraph[0] == "XOR":
        if len(args) != 2:
            raise ValueError("XOR must have exactly two operands")
        return args[0] ^ args[1]
    elif qgraph[0] == "NOT":
        if len(args) != 1:
            raise ValueError("NOT must have exactly one operand")
        return ~args[0]
    raise ValueError(f"Unrecognized operator \"{qgraph[0]}\"")


def assemble_results(qnodes, qedges, **kwargs):
    """Assemble results into Reasoner format."""
    clauses = []

    # assemble result (bindings) and associated (result) kgraph
    node_bindings = [
        (
            "`{0}`: [ni IN collect(DISTINCT `{0}`.id) "
            "WHERE ni IS NOT null "
            "| {{id: ni}}]"
        ).format(
            qnode_id,
        ) if qnode.get("is_set", False) else
        (
            "`{0}`: (CASE "
            "WHEN `{0}` IS NOT NULL THEN [{{id: `{0}`.id{1}}}] "
            "ELSE [] "
            "END)"
        ).format(
            qnode_id,
            f", qnode_id: `{qnode_id}_superclass`.id" if f"{qnode_id}_superclass" in qnodes else "",
        )
        for qnode_id, qnode in qnodes.items()
        if qnode.get("_return", True)
    ]
    edge_bindings = [
        (
            "`{0}`: [ei IN collect(DISTINCT toString(id(`{0}`))) "
            "WHERE ei IS NOT null "
            "| {{id: ei}}]"
        ).format(
            qedge_id,
        ) if kwargs.get("relationship_id", "property") == "internal" else
        (
            "`{0}`: [ei IN collect(DISTINCT `{0}`.id) "
            "WHERE ei IS NOT null "
            "| {{id: ei}}]"
        ).format(
            qedge_id,
        )
        for qedge_id, qedge in qedges.items()
        if qedge.get("_return", True)
    ]
    knodes = [
        "collect(DISTINCT `{0}`)".format(qnode_id)
        for qnode_id, qnode in qnodes.items()
        if qnode.get("_return", True)
    ]
    kedges = [
        "collect(DISTINCT `{0}`)".format(qedge_id)
        for qedge_id, qedge in qedges.items()
        if qedge.get("_return", True)
    ]
    assemble_clause = (
        "WITH {{node_bindings: {{{0}}}, edge_bindings: {{{1}}}}} AS result, "
        "{{nodes: {2}, edges: {3}}} AS knowledge_graph"
    ).format(
        ", ".join(node_bindings) or "",
        ", ".join(edge_bindings) or "",
        " + ".join(knodes) or "[]",
        " + ".join(kedges) or "[]",
    )
    clauses.append(assemble_clause)

    # add SKIP and LIMIT sub-clauses
    clauses.extend(pagination(**kwargs))

    # collect results and aggregate kgraphs
    # also fetch extra knode/kedge properties
    if knodes:
        clauses.append("UNWIND knowledge_graph.nodes AS knode")
    if kedges:
        clauses.append("UNWIND knowledge_graph.edges AS kedge")
    aggregate_clause = "WITH collect(DISTINCT result) AS results, {"
    aggregate_clause += (
        (
            "nodes: apoc.map.fromLists("
            "[n IN collect(DISTINCT knode) | n.id], "
            "[n IN collect(DISTINCT knode) | {"
            "categories: labels(n), name: n.name, "
            "attributes: [key in apoc.coll.subtract(keys(n), "
            + cypher_expression.dumps(RESERVED_NODE_PROPS) +
            ") | {original_attribute_name: key, attribute_type_id: COALESCE("
            + cypher_expression.dumps(ATTRIBUTE_TYPES) +
            "[key], \"NA\"), value: n[key]}]}]), "
        )
        if qnodes else
        "nodes: [], "
    )
    aggregate_clause += (
        (
            "edges: apoc.map.fromLists(" + (
            "[e IN collect(DISTINCT kedge) | toString(ID(e)) ], " if kwargs.get("relationship_id", "property") == "internal" else
            "[e IN collect(DISTINCT kedge) | e.id], "
            ) +
            "[e IN collect(DISTINCT kedge) | {"
            "predicate: type(e), subject: startNode(e).id, object: endNode(e).id, "
            "attributes: [key in apoc.coll.subtract(keys(e), "
            + cypher_expression.dumps(RESERVED_EDGE_PROPS) +
            ") | {original_attribute_name: key, attribute_type_id: COALESCE("
            + cypher_expression.dumps(ATTRIBUTE_TYPES) +
            "[key], \"NA\"), value: e[key]}]}])"
        )
        if kedges else
        "edges: []"
    )
    aggregate_clause += "} AS knowledge_graph"
    clauses.append(aggregate_clause)

    # return results and knowledge graph
    return_clause = "RETURN results, knowledge_graph"
    clauses.append(return_clause)
    return clauses


def pagination(skip=None, limit=None, **kwargs):
    """Get pagination clauses."""
    clauses = []
    if skip is not None:
        clauses.append(f"SKIP {skip}")
    if limit is not None:
        clauses.append(f"LIMIT {limit}")
    return clauses


def get_query(qgraph, **kwargs):
    """Generate a Cypher query to extract the answer maps for a question.

    Returns the query as a string.
    """
    qgraph = copy.deepcopy(qgraph)
    clauses = []
    query = transpile_compound(qgraph, **kwargs)
    clauses.extend(query.compile())
    where_clause = query.where_clause()
    if where_clause:
        if not clauses[-1].startswith("WITH"):
            clauses.append(query.with_clause())
        clauses.append(where_clause)

    if not kwargs.pop("reasoner", True):
        clauses.append(query.return_clause())
        # add SKIP and LIMIT sub-clauses
        clauses.extend(pagination(**kwargs))
    else:
        clauses.extend(assemble_results(
            query.qgraph["nodes"],
            query.qgraph["edges"],
            **kwargs,
        ))

    return " ".join(clauses)



"""MATCHing tools."""
from typing import Dict, List

from bmt import Toolkit

from .nesting import Query
from .util import ensure_list, snake_case, space_case, pascal_case

bmt = Toolkit()


def cypher_prop_string(value):
    """Convert property value to cypher string representation."""
    if isinstance(value, bool):
        return str(value).lower()
    if isinstance(value, str):
        return "\"{0}\"".format(
            value.replace("\"", "\\\"")
        )
    if isinstance(value, (float, int)):
        return str(value)
    raise ValueError(f"Unsupported property type: {type(value).__name__}.")


class NodeReference():
    """Node reference object."""

    def __init__(self, node_id, node, **kwargs):
        """Create a node reference.

        All node properties of types [str, bool, float/int] will be enforced
        verbatim EXCEPT for the following reserved properties:
        * categories
        * ids
        * name
        * is_set
        Un-reserved properties with other types will be coerced to str.
        """
        max_connectivity = kwargs.get("max_connectivity", -1)
        self.anonymous = kwargs.get("anonymous", False)

        node = dict(node)  # shallow copy
        self.name = f"`{node_id}`"

        props = {}
        self._filters = []
        self.labels = []

        category = node.pop("categories", None)
        if isinstance(category, list) and len(category) == 1:
            category = category[0]
        if isinstance(category, list):
            self.labels = ['biolink:NamedThing']
            self._filters.append(" OR ".join([
                "{1} in labels({0})".format(
                    self.name,
                    cypher_prop_string(ci)
                )
                for ci in category
            ]))
        elif category is not None:
            # coerce to a string
            self.labels = [str(category)]

        curie = node.pop("ids", None)
        if isinstance(curie, list) and len(curie) == 1:
            curie = curie[0]
        if isinstance(curie, list):
            self._filters.append("{0}.id in [{1}]".format(
                self.name,
                ", ".join([
                    cypher_prop_string(ci)
                    for ci in curie
                ])
            ))
        elif curie is not None:
            # coerce to a string
            props["id"] = str(curie)

        if max_connectivity > -1:
            self._filters.append("size( ({0})-[]-() ) < {1} + 1".format(
                self.name,
                max_connectivity,
            ))

        props.update(
            (key, value)
            for key, value in node.items()
            if key not in ("name", "is_set", "constraints")
        )

        self.prop_string = "{" + ", ".join([
            f"`{key}`: {cypher_prop_string(value)}"
            for key, value in props.items()
            if value is not None and not key.startswith("_")
        ]) + "}" if props else ""
        self._hints = []
        if curie and self.labels:
            self._hints.append(f"USING INDEX {self.name}:`{self.labels[0]}`(id)")
        self._num = 0

    def __str__(self):
        """Return the cypher node reference."""
        self._num += 1
        if self._num > 1:
            return f"({self.name})"
        ref = "("
        elements = []
        if not self.anonymous:
            elements.append(self.name)
        if self.labels:
            elements.append("".join(f":`{label}`" for label in self.labels))
        elements = ["".join(elements)]
        if self.prop_string:
            elements.append(self.prop_string)
        ref = "({})".format(" ".join(elements))
        return ref

    @property
    def filters(self):
        """Return filters for the cypher node reference.

        To be used in a WHERE clause following the MATCH clause.
        """
        if self._num > 1:
            return []
        return self._filters

    @property
    def hints(self):
        """Return hints for the cypher node reference.

        To be appended to the MATCH clause.
        """
        if self._num > 1:
            return []
        return self._hints


class MissingReference(NodeReference):
    """Missing node reference object."""

    def __init__(self, name):  # pylint: disable=super-init-not-called
        """Initialize."""
        self.name = f"`{name}`"
        self._num = 2


class EdgeReference():
    """Edge reference object."""

    def __init__(
        self,
        edge_id,
        edge,
        anonymous=False,
        invert=True,
    ):
        """Create an edge reference."""
        edge = dict(edge)  # make shallow copy
        _subject = edge.pop("subject")
        _object = edge.pop("object")
        self.name = edge_id if not anonymous else ""
        self.predicates: List[str] = ensure_list(
            edge.pop("predicates", []) or []
        )
        # "related_to" is equivalent to no predicate
        if self.predicates == ["biolink:related_to"]:
            self.predicates = []
        self.filters = []
        self.qualifier_filters = []
        self.label = None  #What goes in the [] on the edge in cypher
        self.length = edge.pop("_length", (1, 1))
        invert = invert and edge.pop("_invert", True)

        self.inverse_predicates = []
        self.directed = False # Controls whether there is an arrow on the edge in cypher
        self.symmetric = True # Whether the original top-level predicates are all symmetric
        self.cypher_invert = False # If true, then the cypher source node will be subject, if false then object
        for predicate in self.predicates:
            el = bmt.get_element(space_case(predicate[8:]))
            if el is None:
                self.symmetric = False #TODO: Is this right?  If there's no predicate isn't it symmetric?
                inverse = None
            else:
                inverse = el.inverse
                self.symmetric = el.symmetric and self.symmetric
            # relationship is directed if any provided predicate is asymmetrical
            if not self.symmetric:
                self.directed = True
            if inverse is not None:
                self.inverse_predicates.append(f"biolink:{snake_case(inverse)}")
            elif self.symmetric:
                self.inverse_predicates.append(predicate)

        # get all descendant predicates
        self.predicates = [
            f"biolink:{snake_case(p)}"
            for predicate in self.predicates
            for p in bmt.get_descendants(space_case(predicate[8:]))
            if bmt.get_element(p).annotations.get('canonical_predicate',False)
        ]
        if not invert:
            self.inverse_predicates = []
        self.inverse_predicates = [
            f"biolink:{snake_case(p)}"
            for predicate in self.inverse_predicates
            for p in bmt.get_descendants(space_case(predicate[8:]))
            if bmt.get_element(p).annotations.get('canonical_predicate', False)
        ]


        unique_preds = list(set(self.predicates + self.inverse_predicates))
        #Having the predicates sorted doesn't matter to neo4j, but it helps in testing b/c we get a consistent string.
        unique_preds.sort()
        self.label = "|".join(
            f"`{predicate}`"
            for predicate in unique_preds
        )

        # We only need the WHERE clause if: we have canonical edges pointing in opposite directions.  In that
        # case we need a non-directed edge and a where clause that points the right ones in different directions
        if not self.symmetric:
            if self.inverse_predicates and self.predicates:
                self.directed = False
                self.filters.append(" OR ".join([
                    "(type(`{0}`) in [{1}] AND startNode(`{0}`) = `{2}`)".format(
                        self.name,
                        ", ".join([
                            f"\"{predicate}\""
                            for predicate in self.predicates
                        ]),
                        _subject,
                    ),
                    "(type(`{0}`) in [{1}] AND startNode(`{0}`) = `{2}`)".format(
                        self.name,
                        ", ".join([
                            f"\"{predicate}\""
                            for predicate in self.inverse_predicates
                        ]),
                        _object,
                    ),
                ]))
            #If we have inverse_predicates and not self.predicates then it means that everything in predicates
            # was non-canonical and we reversed them all so we need to invert the edge
            elif self.inverse_predicates:
                self.cypher_invert=True
        constraints = ["attribute_constraints", "qualifier_constraints"]
        other_non_prop_attributes = ["name", "knowledge_type"]
        self.qualifier_filters = self.__qualifier_filters(edge, edge_id)
        props = {}
        props.update(
            (key, value)
            for key, value in edge.items()
            if key not in set(constraints + other_non_prop_attributes) and not key.startswith("_")
        )

        self.prop_string = " {" + ", ".join([
            f"`{key}`: {cypher_prop_string(value)}"
            for key, value in props.items()
            if value is not None
        ]) + "}" if props else ""

    def __qualifier_filters(self, edge , edge_id):
        constraints = edge.get("qualifier_constraints", [])
        ors = []
        for constraint in constraints:
            ands = []
            for constraint_filter in constraint.get("qualifier_set", []):
                # also handle "qualifier_set": [{ }]
                if not constraint_filter:
                    continue
                all_enums = bmt.get_all_enums()
                qualifiers_values = []
                for enum in all_enums:
                    to_append = bmt.get_enum_value_descendants(enum, constraint_filter['qualifier_value'])
                    if to_append:
                        qualifiers_values += to_append
                # Join qualifier value hierarchy with an or
                qualifier_where_condition =  " ( "+ " OR ".join([f"`{edge_id}`.{constraint_filter['qualifier_type_id']} = {cypher_prop_string(qualifier_value)}"
                                                 for qualifier_value in set(qualifiers_values)]) + " ) "
                ands.append(qualifier_where_condition)
            # if qualifier set is empty ; loop to the next
            if not len(ands):
                continue
            # join contraints in a single qualifier set with `AND`
            ands = ' ( ' + ' AND '.join(ands) + ' ) '
            ors.append(ands)
        # join multiple qualifier sets with `OR`
        return ' OR '.join(ors)

    def __str__(self):
        """Return the cypher edge reference."""
        elements = [
            f":{self.label}" if self.label else "",
        ]
        if not self.length == (1, 1):
            elements.append("*{}..{}".format(
                self.length[0] if self.length[0] is not None else "",
                self.length[1] if self.length[1] is not None else "",
            ))
        elements.append(f"{self.prop_string}")
        return "-[`{0}`{1}]-".format(
            self.name,
            "".join(elements),
        ) + (">" if self.directed else "")


def build_match_clause(
        *patterns,
        filters=None,
        hints=None,
        qualifier_filters=[],
        **kwargs,
):
    """Build MATCH clause (and subclauses) from components."""
    query = ""
    query += "MATCH " + ", ".join(patterns)
    reg_filters_cypher = ""
    qualifier_filters_cypher = ""
    combine_op = ""
    has_filters = False
    if kwargs.get("use_hints", False) and hints:
        query += " " + " ".join(hints)
    if filters or qualifier_filters:
        if len(filters):
            has_filters = True
            if len(filters) > 1:
                filters = [f"({f})" for f in filters]
                reg_filters_cypher += " ( " + " AND ".join(filters) + " ) "
            else:
                reg_filters_cypher += " ( " + filters[0] + " ) "

        if len(qualifier_filters):
            has_filters = True
            qualifier_filters_cypher += f"({qualifier_filters})"

        if len(filters) and len(qualifier_filters):
            combine_op = " AND "
    if has_filters:
        query += " WHERE " + reg_filters_cypher + combine_op + qualifier_filters_cypher

    return query


def match_edge(
    qedge_id,
    qedge,
    node_references: Dict[str, NodeReference],
    invert=True,
    **kwargs,
):
    """Get MATCH clause for edge."""
    eref = EdgeReference(qedge_id, qedge, invert=invert)
    if eref.cypher_invert:
        source_node = node_references[qedge["object"]]
        target_node = node_references[qedge["subject"]]
    else:
        source_node = node_references[qedge["subject"]]
        target_node = node_references[qedge["object"]]
    pattern = f"{source_node}{eref}{target_node}"
    edge_filters = [
        f"({c})"
        for c in source_node.filters + target_node.filters + eref.filters
    ]
    return build_match_clause(
        pattern,
        hints=source_node.hints + target_node.hints,
        filters=edge_filters,
        qualifier_filters=eref.qualifier_filters,
        **kwargs,
    )


def match_query(qgraph, subclass=True, **kwargs):
    """Generate a Cypher MATCH clause.

    Returns the query fragment as a string.
    """
    #If there's no label on a query node, neo4j can't use an index.  This makes such queries basically
    # neo4j killers.   So when there is no category, we add NamedThing which will use an indx.
    for qnode_id, qnode in qgraph["nodes"].items():
        if qnode.get("ids", None) is not None and qnode.get("categories",None) is None:
            qnode["categories"] = ["biolink:NamedThing"]
    if subclass:
        superclasses = {
            qnode_id + "_superclass": {
                "ids": qnode.pop("ids"),
                "categories": qnode.pop("categories", None),
                "_return": False,
            }
            for qnode_id, qnode in qgraph["nodes"].items()
            if qnode.get("ids", None) is not None
        }
        subclass_edges = {
            qnode_id[:-11] + "_subclass_edge": {
                "subject": qnode_id[:-11],
                "object": qnode_id,
                "predicates": ["biolink:subclass_of"],
                "_length": (0, 1),
                "_invert": False,
                "_return": False,
            }
            for qnode_id in superclasses
        }
        qgraph["nodes"].update(superclasses)
        qgraph["edges"].update(subclass_edges)

    # sets of ids
    defined_nodes = set(qgraph["nodes"])
    defined_edges = set(qgraph["edges"])
    referenced_nodes = set(
        [e["subject"] for e in qgraph["edges"].values()]
        + [e["object"] for e in qgraph["edges"].values()]
    )

    # generate internal node and edge variable names
    node_references = {
        qnode_id: NodeReference(qnode_id, qnode, **kwargs)
        for qnode_id, qnode in qgraph["nodes"].items()
    }
    for node_id in referenced_nodes - defined_nodes:  # reference-only nodes
        node_references[node_id] = MissingReference(node_id)

    clauses = []

    # match orphaned nodes
    for node_id in defined_nodes - referenced_nodes:
        clauses.append(build_match_clause(
            str(node_references[node_id]),
            hints=node_references[node_id].hints,
            filters=node_references[node_id].filters,
            **kwargs,
        ))

    # match edges
    for qedge_id, qedge in qgraph["edges"].items():
        clauses.append(match_edge(
            qedge_id,
            qedge,
            node_references,
            **kwargs,
        ))

    return Query(
        " ".join(clauses),
        qids=defined_nodes | defined_edges,
        references=defined_nodes | referenced_nodes | defined_edges,
        qgraph=qgraph,
    )
